class bosko:
    def __init__(self, df, league_season_string, query_position=None):
        import pandas as pd
        if isinstance(df, str):
            df = pd.read_csv(df)
        df['Minutes'] = df['90s'] * 90
        if query_position is not None:
            to_keep = [(query_position == str(pos)[0:2]) for pos in list(df['Pos'])]
            df = df[to_keep]
        self.df = df
        self.league_season_string = league_season_string
        self.models = []
        self.labels = []

    def add_model(self, a, b, model_type, name):
        from balaban.utils import estimate_model
        import numpy as np
        if isinstance(a, str):
            a = self.df[a]
        if isinstance(b, str):
            b = self.df[b]
        a = np.array(a)
        b = np.array(b)
        new_model = estimate_model(a, b, model_type)
        self.models.append(new_model)
        self.labels.append(name)

    def delete_model(self, name):
        import numpy as np
        which_mod = np.min(np.where(self.labels == name)[0])
        del self.models[which_mod]
        del self.labels[which_mod]

    def get_model(self, name):
        import numpy as np
        which_mod = np.min(np.where(np.array(self.labels) == name)[0])
        return self.models[which_mod]

    def make_plot(self, player_query, subtit_text, model_names=None, use_pretty_font=True, dpi=125):
        import matplotlib.pyplot as plt
        import matplotlib.font_manager as fm
        import matplotlib
        import numpy as np
        import os
        from balaban.utils import obtain_player_quantiles
        from importlib_resources import path as get_font_path
        if use_pretty_font:
            font_path = get_font_path('balaban', 'JosefinSans-Regular.ttf')
            try:
                with font_path as f:
                    fname = str(f)
                    fp_title = fm.FontProperties(fname=fname, size=10)
                    fp_subtitle = fm.FontProperties(fname=fname, size=9)
                    fp_labels = fm.FontProperties(fname=fname, size=8)
                    fp_medians = fm.FontProperties(fname=fname, size=7)
            except:
                fp_title = fm.FontProperties(size=10)
                fp_subtitle = fm.FontProperties(size=9)
                fp_labels = fm.FontProperties(size=8)
                fp_medians = fm.FontProperties(size=7)
        else:
            fp_title = fm.FontProperties(size=10)
            fp_subtitle = fm.FontProperties(size=9)
            fp_labels = fm.FontProperties(size=8)
            fp_medians = fm.FontProperties(size=7)

        player_list = np.array(self.df['Player'])
        team_list = np.array(self.df['Squad'])
        mins_played = np.array(self.df['Minutes'])

        if model_names is not None:
            which_mods = np.array([np.min(np.where(self.labels == name)[0]) for name in model_names])
            models = self.models[which_mods]
            labels = self.labels[which_mods]
        else:
            models = self.models
            labels = self.labels

        pl = np.where(player_list == player_query)[0][0]
        n_bars = len(models)
        fig = plt.figure(dpi=dpi)
        ax = plt.subplot(111, projection='polar')
        p90_collected = [None] * n_bars
        for j in range(n_bars):
            h, p90 = obtain_player_quantiles(models[j], pl)
            p90_collected[j] = p90
            ax.bar(np.pi / 2 + j * 2 * np.pi / n_bars,
                   1,
                   color='white',
                   edgecolor='seagreen',
                   width=2 * np.pi / n_bars,
                   linewidth=0.2)
            for i in range(1, 25):
                ax.bar(np.pi / 2 + j * 2 * np.pi / n_bars,
                       h[1][i] - h[1][i - 1],
                       alpha=h[0][i] / np.max(h[0]),
                       color='seagreen',
                       width=2 * np.pi / n_bars,
                       bottom=h[1][i - 1])

        ax.axis('off')
        for quantile in [0.25, 0.5, 0.75]:
            ax.plot(np.linspace(0, 2 * np.pi, 200), quantile * np.ones(200), c='gray',
                    alpha=0.2 + 0.3 * (quantile == 0.5), ls='--')
        ax.plot(np.linspace(0, 2 * np.pi, 200), np.ones(200), c='gray', alpha=0.5)

        fig.suptitle(player_list[pl] + ' (' + team_list[pl] + '), ' + self.league_season_string, y=1.04, fontsize=10,
                     fontproperties=fp_title)
        tit = ax.set_title(subtit_text + '; 90s played: ' + str(mins_played[pl] / 90), fontsize=9, y=1.1,
                           fontproperties=fp_subtitle)
        theta = np.pi / 2 + np.arange(n_bars) * 2 * np.pi / n_bars
        theta_lower = np.pi / 2 + np.arange(n_bars) * 2 * np.pi / n_bars + np.pi / 20
        rotations = np.rad2deg(theta) - 90 - 180 * ((theta > np.pi) & (theta < 2 * np.pi))
        rotations_lower = np.rad2deg(theta_lower) - 90 - 180 * ((theta_lower > np.pi) & (theta_lower < 2 * np.pi))

        for idx in range(n_bars):
            lab = ax.text(theta[idx], 1.2, labels[idx],
                          ha='center',
                          va='center',
                          rotation=rotations[idx],
                          rotation_mode="anchor",
                          fontsize=8,
                          fontproperties=fp_labels, )
            lab = ax.text(theta[idx], 1.075,
                          f'{p90_collected[idx][1]:.3}' + ' (' + f'{p90_collected[idx][0]:.3}' + ', ' + f'{p90_collected[idx][2]:.3}' + ')',
                          ha='center',
                          va='center',
                          rotation=rotations[idx],
                          rotation_mode="anchor",
                          fontsize=9,
                          fontproperties=fp_medians,
                          color='firebrick')
        plt.show()


def scrape_top_five_leagues(path_to_chromedriver, league_names=['epl', 'laliga', 'bundesliga', 'ligue1', 'seriea']):
    from selenium import webdriver
    import pandas as pd
    import numpy as np
    from balaban.utils import get_col_dtype
    from selenium.webdriver.chrome.options import Options

    top_5_league_nums = np.array(['9', '12', '20', '13', '11'])
    top_5_league_names = np.array(['Premier-League', 'La-Liga', 'Bundesliga', 'Ligue-1', 'Serie-A'])
    league_codes = ['epl', 'laliga', 'bundesliga', 'ligue1', 'seriea']
    league_names = list(league_names) if (type(league_names) is np.ndarray) or (
            type(league_names) is tuple) else league_names
    league_names = [league_names] if type(league_names) is str else league_names

    try:
        league_matches = np.isin(league_codes, league_names)
    except:
        raise ValueError(
            "league_names should be a list of characters. Available options are 'epl', 'laliga', 'bundesliga',  "
            "'ligue1', 'seriea'")

    if np.sum(league_matches) == 0:
        raise ValueError(
            "league_names should be a list of characters. Available options are 'epl', 'laliga', 'bundesliga',  "
            "'ligue1', 'seriea'")
    elif np.sum(league_matches) != len(league_names):
        raise ValueError(
            "league_names contains strings that weren't matched. Available options are 'epl', 'laliga', 'bundesliga', "
            " 'ligue1', 'seriea'")

    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument('--headless')
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--disable-dev-shm-usage')
    browser = webdriver.Chrome(path_to_chromedriver, chrome_options=chrome_options)

    top_5_league_names = list(top_5_league_names[league_matches])
    top_5_league_nums = list(top_5_league_nums[league_matches])

    categories = ['passing', 'shooting', 'misc']
    all_players_df = pd.DataFrame()
    for lnum, lnam in zip(top_5_league_nums, top_5_league_names):
        for category in categories:
            url = 'https://fbref.com/en/comps/' + lnum + '/' + category + '/' + lnam + '-Stats'
            browser.get(url)
            if category == 'passing':
                my_table = browser.find_element_by_id('div_stats_passing')
                my_table = my_table.find_element_by_xpath("table")
                df = pd.read_html(my_table.get_attribute('outerHTML'))[0]
                col_names = np.array(
                    [('Unnamed:' not in val[0]) * (val[0] + ': ') + val[1] for val in df.columns.values])
                tmp = np.array(df)[:, np.r_[1, 3, 4, np.arange(7, df.shape[1] - 1)]]
                tmp = np.c_[np.tile(lnam, tmp.shape[0]), tmp]
                tmp_col_names = np.r_[np.array('League'), col_names[np.r_[1, 3, 4, np.arange(7, df.shape[1] - 1)]]]
                league_df = pd.DataFrame(tmp)
                league_df.columns = tmp_col_names
                league_df[league_df['Player'] != 'Player']
                league_df = league_df.astype(league_df.apply(get_col_dtype).to_dict())
                league_df.index = league_df['Player']
            elif category == 'shooting':
                my_table = browser.find_element_by_id('div_stats_shooting')
                my_table = my_table.find_element_by_xpath("table")
                df = pd.read_html(my_table.get_attribute('outerHTML'))[0]
                col_names = df.columns.values
                tmp = np.array(df)[:, np.r_[1, np.arange(8, df.shape[1] - 1)]]
                tmp_df = pd.DataFrame(tmp)
                tmp_df.columns = col_names[np.r_[1, np.arange(8, df.shape[1] - 1)]]
                tmp_df = tmp_df[tmp_df['Player'] != 'Player']
                tmp_df = tmp_df.astype(tmp_df.apply(get_col_dtype).to_dict())
                tmp_df = tmp_df.set_index(list(tmp_df)[0])
                league_df = pd.concat([league_df, tmp_df], axis=1, sort=False)
            else:
                my_table = browser.find_element_by_id('div_stats_misc')
                my_table = my_table.find_element_by_xpath("table")
                df = pd.read_html(my_table.get_attribute('outerHTML'))[0]
                col_names = np.array(
                    [('Unnamed:' not in val[0]) * (val[0] + ': ') + val[1] for val in df.columns.values])
                tmp = np.array(df)[:, np.r_[1, np.arange(8, df.shape[1] - 1)]]
                tmp_df = pd.DataFrame(tmp)
                tmp_df.columns = col_names[np.r_[1, np.arange(8, df.shape[1] - 1)]]
                tmp_df = tmp_df[tmp_df['Player'] != 'Player']
                tmp_df = tmp_df.astype(tmp_df.apply(get_col_dtype).to_dict())
                tmp_df = tmp_df.set_index(list(tmp_df)[0])
                league_df = pd.concat([league_df, tmp_df], axis=1, sort=False)

        all_players_df = pd.concat([all_players_df, league_df])

    return all_players_df



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="balaban",
    version="0.0.16",
    author="Will Thomson",
    author_email="willthomson1991@gmail.com",
    description="Bayesian hierarchical models for football",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/anenglishgoat/balaban",
    packages=setuptools.find_packages(),
    install_requires=['importlib.resources', 'numpy', 'pymc3', 'pandas', 'matplotlib','selenium'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires='>=3.2',
)


def fit_counts_model(counts, mins_played):
    ## estimates a hierarchical poisson model for count data
    ## takes as input:
    ##      counts, a numpy array of shape (num_players,) containing the total numbers of actions completed (across all games)
    ##      mins_played, a numpy array of shape (num_players,) containing the total number of minutes each player was observed for
    ## returns:
    ##      sl, a numpy array of shape (6000,N) containing 6000 posterior samples of actions per 90 (N is the number of players in the
    ##      original data frame who have actually played minutes)
    ##      sb, a numpy array of shape (6000,2) containing 6000 posterior samples of the population-level gamma shape parameter &
    ##                                          the population-level mean
    ##      kk, boolean indicating which players have actually played minutes
    import numpy as np
    import pymc3 as pm
    kk = (mins_played > 0) & np.isfinite(counts)
    mins_played = mins_played[kk]
    counts = counts[kk]
    N = counts.shape[0]

    with pm.Model() as model:
        beta = pm.HalfNormal('beta', sigma=100)
        mu = pm.HalfFlat('mu')
        lambdas = pm.Gamma('lambdas', alpha=mu * beta, beta=beta, shape=N)
        lambda_tilde = lambdas * mins_played
        y = pm.Poisson('y', lambda_tilde, observed=counts)
        approx = pm.fit(n=30000)
    sl = approx.sample(6000)['lambdas'] * 90
    sb = np.c_[approx.sample(6000)['beta'], approx.sample(6000)['mu']]
    return [sl, sb, kk, 'count']


def fit_successes_model(successes, attempts):
    ## estimates a hierarchical binomial model for success rate data
    ## takes as input:
    ##      successes, a numpy array of shape (num_players,) containing the total numbers of successful actions (across all games)
    ##      attempts, a numpy array of shape (num_players,) containing the total numbers of attempted actions (across all games)
    ## returns:
    ##      sl, a numpy array of shape (6000,N) containing 6000 posterior samples of success probabilites (N is the number of players in the
    ##      original data frame who have actually attempted a pass)
    ##      sb, a numpy array of shape (6000,2) containing 6000 posterior samples of the population-level beta parameters
    ##      kk, boolean indicating which players have actually attempted a pass
    import numpy as np
    import pymc3 as pm
    import pymc3.distributions.transforms as tr
    import theano.tensor as tt
    kk = (attempts > 0) & np.isfinite(successes)
    attempts = attempts[kk]
    successes = successes[kk]
    N = attempts.shape[0]

    def logp_ab(value):
        ''' prior density'''
        return tt.log(tt.pow(tt.sum(value), -5 / 2))

    with pm.Model() as model:
        # Uninformative prior for alpha and beta
        ab = pm.HalfFlat('ab',
                         shape=2,
                         testval=np.asarray([1., 1.]))
        pm.Potential('p(a, b)', logp_ab(ab))

        lambdas = pm.Beta('lambdas', alpha=ab[0], beta=ab[1], shape=N)

        p = pm.Binomial('y', p=lambdas, observed=successes, n=attempts)
        approx = pm.fit(n=30000)
    sl = approx.sample(6000)['lambdas'] * 100
    sb = approx.sample(6000)['ab']
    return [sl, sb, kk, 'success']


def fit_expected_successes_per_action_model(xS, attempts):
    ## estimates a hierarchical binomial model for success rate data
    ## takes as input:
    ##      sp, a numpy array of shape (num_players,) containing the expected successes per action for each player (e.g. xG per shot, xA per KP)
    ##      attempts, a numpy array of shape (num_players,) containing the total numbers of attempted actions for each player (e.g. shots, key passes)
    ## returns:
    ##      sl, a numpy array of shape (6000,N) containing 6000 posterior samples of success probabilites (N is the number of players in the
    ##      original data frame who have registered non-zero expected succcesses)
    ##      sb, a numpy array of shape (6000,3) containing 6000 posterior samples of: the population-level & observation-level beta 'sample size'
    ##              parameters and the population-level mean
    ##      kk, boolean indicating which players have actually registered non-zero expected successes
    import numpy as np
    import pymc3 as pm
    kk = (attempts > 0) & (xS > 0)
    sp = xS[kk] / attempts[kk]
    attempts = attempts[kk]
    N = attempts.shape[0]

    with pm.Model() as model:
        v = pm.HalfNormal('v', shape=2, sigma=100)
        mu = pm.Uniform('mu')
        lambdas = pm.Beta('lambdas', alpha=mu * v[0], beta=(1 - mu) * v[0], shape=N)
        y = pm.Beta('y',
                    alpha=lambdas * (attempts * (v[1] + 1) - 1),
                    beta=(1 - lambdas) * (attempts * (v[1] + 1) - 1),
                    observed=sp)
        approx = pm.fit(n=30000)
    sl = approx.sample(6000)['lambdas']
    sb = np.c_[approx.sample(6000)['v'], approx.sample(6000)['mu']]
    return [sl, sb, kk, 'expected']


def fit_expected_successes_per90_model(xSuccess_model, attempts_model):
    ## the inputs are two models which should have been returned by:
    ##    fit_expected_successes_per_action_model (first argument)
    ##    fit_counts_model (second argument)
    ## the input models should estimate
    ##    the number of actions attempted per 90 (e.g. shots or key passes) -- fit on count data
    ##    the probability per action that they lead to the corresponding desired outcome (e.g. goal or assist) -- fit on xG/xA data
    kk = (xSuccess_model[2] & attempts_model[2])
    sl = (attempts_model[0][:, kk[attempts_model[2]]] * xSuccess_model[0][:, kk[xSuccess_model[2]]])
    return [sl, [], kk, 'expected_per90']


def fit_adj_pass_model(successes, attempts):
    ## inputs are two lists in the form:
    ##       successes = [successful long passes, total successful passes]
    ##       attempts = [attempted long passes, total attempted passes]
    ## returns:
    ##      sl, a numpy array of shape (6000,N) containing 6000 posterior samples of success probabilites (N is the number of players in the
    ##      original data frame who have registered non-zero expected succcesses)
    ##      sb, an empty list
    ##      kk, boolean indicating which players have actually registered non-zero expected successes
    ##      'adj_pass', character string indicating the model type.
    import numpy as np
    import pymc3 as pm
    import pymc3.distributions.transforms as tr
    import theano.tensor as tt
    LonCmp = successes[0];
    TotCmp = successes[1]
    LonAtt = attempts[0];
    TotAtt = attempts[1]
    kk = (LonCmp > 0) & np.isfinite(LonAtt)
    LonCmp = LonCmp[kk];
    LonAtt = LonAtt[kk];
    TotCmp = TotCmp[kk];
    TotAtt = TotAtt[kk]
    ShCmp = TotCmp - LonCmp;
    ShAtt = TotAtt - LonAtt
    average_long_tendency = np.mean(LonAtt / TotAtt)
    N = np.sum(kk)

    def logp_ab(value):
        ''' prior density'''
        return tt.log(tt.pow(tt.sum(value), -5 / 2))

    with pm.Model() as model:
        # Uninformative prior for alpha and beta
        ab_short = pm.HalfFlat('ab_short',
                               shape=2,
                               testval=np.asarray([1., 1.]))
        ab_long = pm.HalfFlat('ab_long',
                              shape=2,
                              testval=np.asarray([1., 1.]))
        pm.Potential('p(a_s, b_s)', logp_ab(ab_short))
        pm.Potential('p(a_l, b_l)', logp_ab(ab_long))

        lambda_short = pm.Beta('lambda_s', alpha=ab_short[0], beta=ab_short[1], shape=N)
        lambda_long = pm.Beta('lambda_l', alpha=ab_long[0], beta=ab_long[1], shape=N)

        y_short = pm.Binomial('y_s', p=lambda_short, observed=ShCmp, n=ShAtt)
        y_long = pm.Binomial('y_l', p=lambda_short * lambda_long, observed=LonCmp, n=LonAtt)
        approx = pm.fit(n=30000)
    s_sh = approx.sample(6000)['lambda_s']
    s_lo = approx.sample(6000)['lambda_l']
    sl = average_long_tendency * s_lo + (1 - average_long_tendency) * s_sh
    return [sl, [], kk, 'adj_pass']


def estimate_model(a, b, model_type):
    if model_type == 'count':
        out = fit_counts_model(a, b)
    elif model_type == 'success_rate':
        out = fit_successes_model(a, b)
    elif model_type == 'xSpA':
        out = fit_expected_successes_per_action_model(a, b)
    elif model_type == 'adj_pass':
        try:
            out = fit_adj_pass_model(a, b)
        except ValueError:
            print(
                "Check inputs. The inputs should be two lists of the form [successful long passes, total successful passes] & [attempted long passes, total attempted passes]")
    elif model_type == 'xSp90':
        try:
            out = fit_expected_successes_per90_model(a, b)
        except ValueError:
            print(
                "Check inputs. The inputs should be two pre-estimated models. The first argument should be a list returned by a 'counts' model. The second argument should be a list returned by an 'xSpA' model.")
    else:
        raise ValueError("Invalid model_type. model_type should be one of 'count', 'success_rate', 'xSpA', or 'xSp90'")
    return out


def obtain_player_quantiles(model, player_index):
    model_type = model[3]
    import numpy as np
    if model_type == 'count':
        from scipy.stats import gamma
        pind = np.arange(np.shape(model[2])[0])[model[2]]
        pind = np.where(pind == player_index)[0]
        percentile_hist = np.histogram(
            gamma.cdf(model[0][:, pind],
                      a=np.mean(model[1][:, 1]) * np.mean(model[1][:, 0]),
                      scale=90 / np.mean(model[1][:, 0])),
            bins=25)
        per90_quantiles = np.quantile(model[0][:, pind], [0.125, 0.5, 0.875])
    elif model_type == 'success':
        from scipy.stats import beta
        pind = np.arange(np.shape(model[2])[0])[model[2]]
        pind = np.where(pind == player_index)[0]
        percentile_hist = np.histogram(
            beta.cdf(model[0][:, pind] / 100,
                     a=np.mean(model[1][:, 0]),
                     b=np.mean(model[1][:, 1])),
            bins=25)
        per90_quantiles = np.quantile(model[0][:, pind], [0.05, 0.5, 0.95])
    elif model_type == 'expected':
        from scipy.stats import beta
        pind = np.arange(np.shape(model[2])[0])[model[2]]
        pind = np.where(pind == player_index)[0]
        percentile_hist = np.histogram(
            beta.cdf(model[0][:, pind],
                     a=np.mean(model[1][:, 2]) * np.mean(model[1][:, 0]),
                     b=(1 - np.mean(model[1][:, 2])) * np.mean(model[1][:, 0])),
            bins=25)
        per90_quantiles = np.quantile(model[0][:, pind], [0.05, 0.5, 0.95])
    elif (model_type == 'expected_per90') | (model_type == 'adj_pass'):
        pind = np.arange(np.shape(model[2])[0])[model[2]]
        pind = np.where(pind == player_index)[0]
        samps_flat = np.random.choice(model[0].flatten(), size=5000)  ## downsample to make cdf quicker to compute
        sorted_samps = np.sort(samps_flat)
        ecdf = lambda x: np.sum(sorted_samps[:, None] < x, axis=0) / len(sorted_samps)
        percentile_hist = np.histogram(
            ecdf(model[0][:, pind].T),
            bins=25)
        per90_quantiles = np.quantile(model[0][:, pind], [0.05, 0.5, 0.95])
    else:
        print("Invalid model type. Must be one of 'count', 'success', 'expected', 'expected_per90' or 'adj_pass'")
    return percentile_hist, per90_quantiles


def get_col_dtype(col):
    """
    Infer datatype of a pandas column, process only if the column dtype is object.
    input:   col: a pandas Series representing a df column.
    """

    if col.dtype == "object":
        try:
            col_new = pd.to_datetime(col.dropna().unique())
            return col_new.dtype
        except:
            try:
                col_new = pd.to_numeric(col.dropna().unique())
                return col_new.dtype
            except:
                try:
                    col_new = pd.to_timedelta(col.dropna().unique())
                    return col_new.dtype
                except:
                    return "object"

    else:
        return col.dtype
